# frozen_string_literal: true

module Landline
  # Shared DSL methods
  module DSL
    # Path (and subclasses) DSL constructors
    module PathConstructors
      # (in Landline::Path context)
      # Append a Node child object to the list of children
      def register(obj)
        unless obj.is_a? Landline::Node
          raise ArgumentError, "register accepts node children only"
        end

        @origin.children.append(obj)
      end

      # (in Landline::Path context)
      # Create a new {Landline::Path} object
      def path(path, **args, &setup)
        register(Landline::Path.new(path, parent: @origin, **args, &setup))
      end

      # (in Landline::Path context)
      # Create a new {Landline::Handlers::Probe} object
      def probe(path, **args, &_setup)
        register(Landline::Handlers::Probe.new(path,
                                               parent: @origin,
                                               **args))
      end

      # (in Landline::Path context)
      # Create a new {Landline::Handlers::GETHandler} object
      def get(path, **args, &setup)
        register(Landline::Handlers::GET.new(path,
                                             parent: @origin,
                                             **args,
                                             &setup))
      end

      # (in Landline::Path context)
      # create a new {Landline::Handlers::POSTHandler} object
      def post(path, **args, &setup)
        register(Landline::Handlers::POST.new(path,
                                              parent: @origin,
                                              **args,
                                              &setup))
      end

      # (in Landline::Path context)
      # Create a new {Landline::Handlers::PUTHandler} object
      def put(path, **args, &setup)
        register(Landline::Handlers::PUT.new(path,
                                             parent: @origin,
                                             **args,
                                             &setup))
      end

      # (in Landline::Path context)
      # Create a new {Landline::Handlers::HEADHandler} object
      def head(path, **args, &setup)
        register(Landline::Handlers::HEAD.new(path,
                                              parent: @origin,
                                              **args,
                                              &setup))
      end

      # (in Landline::Path context)
      # Create a new {Landline::Handlers::DELETEHandler} object
      def delete(path, **args, &setup)
        register(Landline::Handlers::DELETE.new(path,
                                                parent: @origin,
                                                **args,
                                                &setup))
      end

      # (in Landline::Path context)
      # Create a new {Landline::Handlers::CONNECTHandler} object
      def connect(path, **args, &setup)
        register(Landline::Handlers::CONNECT.new(path,
                                                 parent: @origin,
                                                 **args,
                                                 &setup))
      end

      # (in Landline::Path context)
      # Create a new {Landline::Handlers::TRACEHandler} object
      def trace(path, **args, &setup)
        register(Landline::Handlers::TRACE.new(path,
                                               parent: @origin,
                                               **args,
                                               &setup))
      end

      # (in Landline::Path context)
      # Create a new {Landline::Handlers::PATCHHandler} object
      def patch(path, **args, &setup)
        register(Landline::Handlers::PATCH.new(path,
                                               parent: @origin,
                                               **args,
                                               &setup))
      end

      # (in Landline::Path context)
      # Create a new {Landline::Handlers::OPTIONSHandler} object
      def options(path, **args, &setup)
        register(Landline::Handlers::OPTIONS.new(path,
                                                 parent: @origin,
                                                 **args,
                                                 &setup))
      end

      # (in Landline::Path context)
      # Create a new {Landline::Handlers::GETHandler} that serves static files
      def serve(path)
        register(Landline::Handlers::Serve.new(path, parent: @origin))
      end

      # (in Landline::Path context)
      # Create a new application crosscall link (acts like #call in probe context and strips its path from request)
      def link(path, application)
        register(Landline::Handlers::Link.new(path,
                                              application,
                                              parent: @origin))
      end
    end
  end
end
