# frozen_string_literal: true

module Landline
  # Shared DSL methods
  module DSL
    # Common path methods
    module PathMethods
      # (in Landline::Path context)
      # Bounce request if no handler found instead of issuing 404
      def bounce
        @origin.bounce = true
      end

      # (in Landline::Path context)
      # Unset bounce
      def nobounce
        @origin.bounce = false
      end

      # (in Landline::Path context)
      # Create a status code handler on path.
      # Recursively applies to all paths unless overridden.
      # @param code [Integer, nil] Specify a status code to handle
      # @param block [#call] Block to run on given code
      def handle(code = nil, &block)
        @origin.properties["handle.#{code || 'default'}"] = block
      end

      # (in Landline::Path context)
      # Insert a pass-through pipeline into request processing
      # (i.e. for error handling purposes).
      # Passed block should yield request (and return yielded data back).
      # @param block [#call] block that yields request
      def pipeline(&block)
        @origin.pipeline = block
      end

      # (in Landline::Path context)
      # Set path index
      # @param index [Array,String]
      def index(index)
        case index
        when Array
          @origin.properties['index'] = index
        when String
          @origin.properties['index'] = [index]
        else
          raise ArgumentError, "index should be an Array or a String"
        end
      end

      # (in Landline::Path context)
      # Set root path (appends matched part of the path).
      # @param path [String]
      def root(path)
        @origin.root = File.expand_path(path)
      end

      # (in Landline::Path context)
      # Set root path (without appending matched part).
      # @param path [String]
      def remap(path)
        @origin.remap = File.expand_path(path)
      end

      # (in Landline::Path context)
      # Add a preprocessor to the path.
      # Does not modify path execution.
      # @param block [#call]
      # @yieldparam request [Landline::Request]
      def preprocess(&block)
        @origin.preprocess(&block)
        block
      end

      # (in Landline::Path context)
      # Add a postprocessor to the path.
      # @param block [#call]
      # @yieldparam request [Landline::Request]
      # @yieldparam response [Landline::Response]
      def postprocess(&block)
        @origin.postprocess(&block)
        block
      end

      alias before preprocess
      alias after postprocess

      # (in Landline::Path context)
      # Add a filter to the path.
      # Blocks path access if a filter returns false.
      # @param block [#call]
      # @yieldparam request [Landline::Request]
      def filter(&block)
        @origin.filter(&block)
        block
      end

      # (in Landline::Path context)
      # Include an application as a child of path.
      # @deprecated this method is being deprecated due to strong dependency on the framework
      # @param filename [String]
      def plugin(filename)
        define_singleton_method(:run) do |object|
          unless object.is_a? Landline::Node
            raise ArgumentError, "not a node instance or subclass instance"
          end

          object
        end
        @origin.children.append(
          instance_eval(File.read(filename), filename)
        )
        singleton_class.undef_method :run
      end
    end
  end
end
