# frozen_string_literal: true

require_relative '../response'
require_relative '../util/multipart'
require_relative '../util/parseutils'
require_relative '../util/html'
require 'json'

module Landline
  module DSL
    # Common methods for Probe objects
    module ProbeMethods
      # (in Landline::Probe context)
      # Get the current request
      # @return [Landline::Request]
      def request
        @origin.request
      end

      # (in Landline::Probe context)
      # Set response status (generate response if one doesn't exist yet)
      # @param status [Integer] http status code
      def status(status)
        @origin.response = (@origin.response or Landline::Response.new)
        @origin.response.status = status
      end

      # (in Landline::Probe context)
      # Add a finalizer callable to the response
      # @param callable [#call]
      def defer(&callable)
        rack = @origin.request.rack
        if rack.respond_to?(:response_finished)
          rack.response_finished.append(callable)
        end
        # puma for some reason isn't compatible with the 3.0.0 spec on this
        rack.after_reply.append(callable) if rack.respond_to?(:after_reply)
      end

      # (in Landline::Probe context)
      # Do serverside request redirection
      # @note this essentially reprocesses the whole request - be mindful of processing time!
      # @param path [String]
      def jump(path)
        @origin.request.path = path
        throw(:break, [307, { "x-internal-jump": true }, []])
      end

      # (in Landline::Probe context)
      # Do clientside request redirection via 302 code
      # @param path [String]
      def redirect(path)
        throw(:break, [302, { "location": path }, []])
      end

      # (in Landline::Probe context)
      # Do clientside request redirection via 307 code
      # @param path [String]
      def redirect_with_method(path)
        throw(:break, [307, { "location": path }, []])
      end

      alias code status

      # (in Landline::Probe context)
      # Set a partial hijack callback
      # @param block [#call] Callable block
      def partial_hijack(&block)
        @origin.response ||= Landline::Response.new
        @origin.response.add_header("rack.hijack", block)
      end

      # (in Landline::Probe context)
      # Fully hijack IO
      # @return [IO]
      def hijack
        @origin.request.hijack.call
      end

      # (in Landline::Probe context)
      # Set response header (generate response if one doesn't exist yet)
      # @param key [String] header name
      # @param value [String] header value
      def header(key, value)
        return status(value) if key.downcase == "status"

        unless key.match(Landline::Util::HeaderRegexp::TOKEN)
          raise ArgumentError, "header key has invalid characters"
        end

        unless value&.match(Landline::Util::HeaderRegexp::PRINTABLE)
          raise ArgumentError, "value key has invalid characters"
        end

        @origin.response = (@origin.response or Landline::Response.new)
        key = key.downcase.to_s
        @origin.response.add_header(key, value)
      end

      # (in Landline::Probe context)
      # Delete a header value from the headers hash
      # If no value is provided, deletes all key entries
      # @param key [String] header name
      # @param value [String, nil] header value
      def delete_header(key, value = nil)
        return unless @origin.response

        return if key.downcase == "status"

        unless key.match(Landline::Util::HeaderRegexp::TOKEN)
          raise ArgumentError, "header key has invalid characters"
        end

        unless value&.match(Landline::Util::HeaderRegexp::PRINTABLE)
          raise ArgumentError, "value key has invalid characters"
        end

        @origin.response.delete_header(key.to_s, value)
      end

      # (in Landline::Probe context)
      # Set response cookie
      # @see Landline::Cookie.new
      def cookie(*params, **options)
        @origin.response = (@origin.response or Landline::Response.new)
        @origin.response.add_cookie(
          Landline::Cookie.new(*params, **options)
        )
      end

      # (in Landline::Probe context)
      # Delete a cookie
      # If no value is provided, deletes all cookies with the same key
      # @param key [String] cookie key
      # @param value [String, nil] cookie.value
      def delete_cookie(key, value = nil)
        return unless @origin.response

        @origin.response.delete_cookie(key, value)
      end

      # (in Landline::Probe context)
      # Checks if current request has multipart/form-data associated with it
      # @return [Boolean]
      def form?
        value, opts = _verify_content_type('multipart/form-data')
        !!(value && opts && opts['boundary'])
      end

      # (in Landline::Probe context)
      # Returns formdata
      # @note reads request.input - may nullify request.body.
      # @return [Hash{String=>(String,Landline::Util::FormPart)}]
      def form
        _, opts = Landline::Util::ParserCommon.parse_value(
          request.headers["content-type"]
        )
        Landline::Util::MultipartParser.new(
          request.input, opts["boundary"]
        ).to_h
      end

      # (in Landline::Probe context)
      # Checks if current request has urlencoded query string
      # @return [Boolean]
      def query?
        !!_verify_content_type("application/x-www-form-urlencode")
      end

      # (in Landline::Probe context)
      # Returns parsed query hash
      # @note reads request.body - may nullify .input, .body data is memoized
      # @return [Hash{String => Object}] query data
      def query
        Landline::Util::Query.new(request.body).parse
      end

      # (in Landline::Probe context)
      # Returns shallow parsed query hash
      # @note reads request.body - may nullify .input, .body data is memoized
      # @return [Hash{String => Object}] query data
      def query_shallow
        Landline::Util::Query.new(request.body).parse_shallow
      end

      # (in Landline::Probe context)
      # Check if body is a JSON object
      # @return [Boolean]
      def json?
        !!_verify_content_type('application/json')
      end

      # (in Landline::Probe context)
      # Return parse JSON object
      # @note reads request.input - may nullify request.body.
      # @return [Object]
      def json
        JSON.parse(request.input)
      end

      # (in Landline::Probe context)
      # Open a file relative to current filepath
      # @see File.open
      def file(path, mode = "r", *all, &block)
        File.open("#{request.filepath}/#{path}", mode, *all, &block)
      end

      # (in Landline::Probe context)
      # Escape HTML entities
      # @see Landline::Util.escape_html
      def escape_html(text)
        Landline::Util.escape_html(text)
      end

      # (in Landline::Probe context)
      # Unescape HTML entities
      # @see Landline::Util.escape_html
      def unescape_html(text)
        Landline::Util.unescape_html(text)
      end

      # (in Landline::Path context)
      # Pass the requested environment to a different application
      # @param application [#call] Rack application
      # @return [Array(Integer, Hash{String => Object}, Object)] response
      def call(application)
        application.call(@origin.request.env)
      end

      private

      def _verify_content_type(type)
        return false unless request.headers['content-type']

        value, opts = Landline::Util::ParserCommon.parse_value(
          request.headers["content-type"]
        )
        if value == type and
           request.input
          [value, opts]
        else
          false
        end
      end
    end
  end
end
