# frozen_string_literal: true

require 'securerandom'

module Landline
  # Module for controlling session signing secrets
  module Session
    # Set hmac secret
    # @param secret [String]
    def self.hmac_secret=(secret)
      @hmac_secret = secret
    end

    # Get hmac secret
    def self.hmac_secret
      unless @hmac_secret or ENV['HMAC_SECRET']
        warn <<~MSG
          warn: hmac secret not supplied, using randomized one
          warn: provide hmac secret with $HMAC_SECRET or Landline::Session.hmac_secret
        MSG
      end
      @hmac_secret ||= ENV.fetch('HMAC_SECRET', SecureRandom.base64(80))
    end

    # Class for representing session errors
    class SessionError < ::StandardError
    end

    # Class for representing session storage
    class Session
      def initialize(cookie, cookies_callback)
        @data = if cookie
                  Landline::Util::JWT.from_string(
                    cookie,
                    Landline::Session.hmac_secret
                  )
                else
                  Landline::Util::JWT.new({})
                end
        @valid = !@data.nil?
        @cookies_callback = cookies_callback
      end

      # Retrieve data from session storage
      # @param key [String, Symbol] serializable key
      def [](key)
        raise Landline::Session::SessionError, "session not valid" unless @valid

        unless key.is_a? String or key.is_a? Symbol
          raise StandardError, "key not serializable"
        end

        @data.data[key]
      end

      # Set data to session storage
      # @param key [String, Symbol] serializable key
      # @param value [Object] serializable data
      def []=(key, value)
        raise Landline::Session::SessionError, "session not valid" unless @valid

        unless key.is_a? String or key.is_a? Symbol
          raise StandardError, "key not serializable"
        end

        @data.data[key] = value
        @cookies_callback.call(@data.make(Landline::Session.hmac_secret))
      end

      attr_reader :valid
    end
  end
end

module Landline
  module DSL
    module ProbeMethods
      # TODO: If probe execution contexts are somehow shared between threads
      # this could result in a session leakage through race condition.

      # (in Landline::Probe context)
      # Return session storage hash
      # @return [Landline::Session::Session]
      def session
        return @session if @session

        @session = Landline::Session::Session.new(
          request.cookies.dig('session', 0)&.value,
          proc do |value|
            delete_cookie("session", value)
            cookie("session", value)
          end
        )
        request.postprocessors.append(proc do
          @session = nil
        end)
        @session
      end
    end
  end
end
