# frozen_string_literal: true

require_relative 'pattern_matching/util'

module Landline
  # Abstract class that reacts to request navigation.
  # Does nothing by default, behaviour should be overriden through
  # #reject and #process
  # @abstract
  class Node
    # @param path [Object]
    # @param filepath [Boolean] should Node modify request.filepath.
    def initialize(path, parent:, filepath: true)
      @pattern = Pattern.new(path).freeze
      @properties = Landline::Util::Lookup.new(parent&.properties)
      @root = nil
      @remap = false
      @modify_filepath = filepath
    end

    # Set Node file root (like root in Nginx)
    # @param path [String]
    def root=(path)
      raise ArgumentError, "path should be a String" unless path.is_a? String

      @properties["path"] = File.expand_path(path)
      @root = File.expand_path(path)
    end

    # Set Node absolute file path (like alias in Nginx)
    # @param path [String]
    def remap=(path)
      self.root = path
      @remap = true
    end

    # Try to navigate the path. Run method callback in response.
    # @param request [Landline::Request]
    # @return [Boolean]
    def go(request)
      # rejected at pattern
      return reject(request) unless @pattern.match?(request.path)

      request.push_state
      path, splat, param = @pattern.match(request.path)
      do_filepath(request, request.path.delete_suffix('/')
                                       .delete_suffix(path))
      request.path = path
      request.splat.append(*splat)
      request.param.merge!(param)
      value = process(request)
      # rejected at callback - restore state
      request.pop_state unless value
      # finally, return process value
      value
    end

    # Method callback on failed request navigation
    # @param _request [Landline::Request]
    # @return false
    def reject(_request)
      false
    end

    # Method callback on successful request navigation
    # @param _request [Landline::Request]
    # @return true
    def process(_request)
      true
    end

    attr_reader :remap, :root

    private

    # Process filepath for request
    def do_filepath(request, path)
      return unless @modify_filepath

      if @root
        request.filepath = "#{@root}/#{@remap ? '' : path}/"
      else
        request.filepath += "/#{path}/"
      end
      request.filepath.gsub!(/\/+/, "/")
    end
  end
end
