# frozen_string_literal: true

require_relative 'node'
require_relative 'util/lookup'
require_relative 'dsl/methods_probe'
require_relative 'dsl/methods_common'
require_relative 'dsl/constructors_probe'
require 'pp'

module Landline
  # All subclasses of Landline::Probe that respond to requests
  module Handlers
    autoload :Handler, "landline/probe/handler"
    autoload :GET, "landline/probe/http_method"
    autoload :POST, "landline/probe/http_method"
    autoload :HEAD, "landline/probe/http_method"
    autoload :PUT, "landline/probe/http_method"
    autoload :DELETE, "landline/probe/http_method"
    autoload :CONNECT, "landline/probe/http_method"
    autoload :OPTIONS, "landline/probe/http_method"
    autoload :TRACE, "landline/probe/http_method"
    autoload :PATCH, "landline/probe/http_method"
    autoload :Serve, "landline/probe/serve_handler"
    autoload :Link, "landline/probe/crosscall_handler"
  end

  # Context that provides execution context for Probes.
  class ProbeContext
    include Landline::DSL::ProbeConstructors
    include Landline::DSL::ProbeMethods
    include Landline::DSL::CommonMethods

    def initialize(origin)
      @origin = origin
    end
  end

  # Ephemeral proxy class to which callback execution binds
  class ProbeExecutionOrigin
    def initialize(request, properties)
      @request = request
      @properties = Landline::Util::LookupROProxy.new(properties)
    end

    attr_accessor :response
    attr_reader :request, :properties
  end

  # Test probe. Also base for all "reactive" nodes.
  class Probe < Landline::Node
    # @param path [Object]
    # @param parent [Landline::Node]
    def initialize(path, parent:, **params)
      super(path, parent: parent, **params)
    end

    attr_reader :properties

    # Method callback on successful request navigation.
    # Throws an error upon reaching the path.
    # This behaviour should only be used internally.
    # @param request [Landline::Request]
    # @return [Boolean] true if further navigation is possible
    # @raise [StandardError]
    def process(request)
      return reject(request) unless request.path.match?(/^\/?$/)

      raise StandardError, <<~STREND
        probe reached #{request.splat.inspect}, #{request.param.inspect}
        #{request.pretty_inspect}
      STREND
    end
  end
end
