# frozen_string_literal: true

require_relative "../probe"

module Landline
  module Handlers
    # Probe that executes a callback on request
    class Handler < Landline::Probe
      # @param path [Object]
      # @param args [Hash] hashed parameters to passthrough to Probe
      # @param exec [#call]
      def initialize(path, **args, &exec)
        super(path, **args)
        @callback = exec
      end

      attr_accessor :response
      attr_reader :request

      # Method callback on successful request navigation.
      # Runs block supplied with object initialization.
      # Request's #splat and #param are passed to block.
      #
      # Callback's returned should be one of viable responses:
      #
      # - {Landline::Response} object
      # - An array that matches Rack return form
      # - An array that matches old (Rack 2.x) return form
      # - A string (returned as HTML with code 200)
      # - false (bounces the request to next handler)
      # @param request [Landline::Request]
      # @return [Boolean] true if further navigation is possible
      # @raise [UncaughtThrowError] may raise if die() is called.
      def process(request)
        origin = Landline::ProbeExecutionOrigin.new(request, @properties)
        context = Landline::ProbeContext.new(origin)
        return reject(request) unless request.path.match?(/^\/?$/)

        response = catch(:break) do
          context.instance_exec(*request.splat,
                                **request.param,
                                &@callback)
        end
        return false unless response

        oresponse = origin.response
        if oresponse and [String, File, IO].include? response.class
          oresponse.body = response
          throw :finish, oresponse
        end
        throw :finish, response
      end
    end
  end
end
