# frozen_string_literal: true

require 'uri'
require_relative 'util/query'
require_relative 'util/cookie'

module Landline
  # Request wrapper for Rack protocol
  class Request
    # @param env [Array]
    def initialize(env)
      # Should not be used under regular circumstances or depended upon.
      @_original_env = env
      # Rack environment variable bindings. Should be public and frozen.
      init_request_params(env)
      # Cookie hash
      @cookies = Landline::Cookie.from_cookie_string(@headers['cookie'])
      # Query parsing
      @query = Landline::Util::Query.new(@query_string)
      # Pattern matching parameters. Public, readable, unfrozen.
      @param = {}
      @splat = []
      # Traversal route. Public and writable.
      @path = URI.decode_www_form_component(env["PATH_INFO"])
      # File serving path. Public and writable.
      @filepath = "/"
      # Encapsulates all rack variables. Is no longer private, but usually should not be used directly
      @rack = init_rack_vars(env)
      # Internal navigation states. Private.
      @states = []
      # Postprocessors for current request
      @postprocessors = []
    end

    # Run postprocessors
    # @param response [Landline::Response]
    def run_postprocessors(response)
      @postprocessors.reverse_each do |postproc|
        postproc.call(self, response)
      end
      @postprocessors = []
    end

    # Returns request body (if POST data exists)
    # @note reads data from rack.input, which is not rewindable. .body data is memoized.
    # @return [nil, String]
    def body
      @body ||= @rack.input&.read
    end

    # Returns raw Rack input object
    # @note Rack IO is not always rewindable - if it is read once, the data is gone (i.e. request.body will return nothing).
    # @return [IO] (May not entirely be compatible with IO, see Rack/SPEC.rdoc)
    def input
      @rack.input
    end

    # Push current navigation state (path, splat, param) onto state stack
    def push_state
      @states.push([@path, @param.dup, @splat.dup, @filepath.dup])
    end

    # Load last navigation state (path, splat, param) from state stack
    def pop_state
      @path, @param, @splat, @filepath = @states.pop
    end

    # Checks if response stream can be partially hijacked
    def hijack?
      @_original_env['rack.hijack?']
    end

    # Returns full hijack callback
    def hijack
      @_original_env['rack.hijack']
    end

    # Reconstructs rack env after modification
    def env
      path = @path
      @_original_env.merge(reconstruct_headers)
                    .merge({
                             'PATH_INFO' => path,
                             'REQUEST_PATH' => path,
                             'QUERY_STRING' => query.query,
                             'REQUEST_URI' => "#{path}?#{query.query}"
                           })
                    .merge(reconstruct_cookie)
    end

    attr_reader :request_method, :script_name, :path_info, :server_name,
                :server_port, :server_protocol, :headers, :param, :splat,
                :postprocessors, :cookies, :rack
    attr_accessor :path, :filepath, :query

    private

    # Initialize basic rack request parameters
    # @param env [Hash]
    def init_request_params(env)
      @request_method = env["REQUEST_METHOD"]
      @script_name = env["SCRIPT_NAME"]
      @path_info = env["PATH_INFO"]
      @query_string = env["QUERY_STRING"]
      @server_name = env["SERVER_NAME"]
      @server_port = env["SERVER_PORT"]
      @server_protocol = env["SERVER_PROTOCOL"]
      @headers = init_headers(env)
    end

    # Initialize rack parameters struct
    # @param env [Hash]
    # @return Object
    def init_rack_vars(env)
      rack_vars = env.filter_map do |k, v|
        [k.delete_prefix("rack."), v] if k.start_with? "rack."
      end.to_h
      return if rack_vars.empty?

      rack_vars["multipart"] = init_multipart_vars(env)
      rack_keys = rack_vars.keys
      rack_keys_sym = rack_keys.map(&:to_sym)
      Struct.new(*rack_keys_sym)
            .new(*rack_vars.values_at(*rack_keys))
            .freeze
    end

    # Initialize multipart parameters struct
    # @param env [Hash]
    # @return Object
    def init_multipart_vars(env)
      multipart_vars = env.filter_map do |k, v|
        if k.start_with? "rack.multipart"
          [k.delete_prefix("rack.multipart."), v]
        end
      end.to_h
      return if multipart_vars.empty?

      multipart_keys = multipart_vars.keys
      multipart_keys_sym = multipart_keys.map(&:to_sym)
      Struct.new(*multipart_keys_sym)
            .new(*multipart_vars.values_at(*multipart_keys))
            .freeze
    end

    # Initialize headers hash
    # @param env [Hash]
    # @return Hash
    def init_headers(env)
      headers = env.filter_map do |name, value|
        [name.delete_prefix("HTTP_"), value] if name.start_with?("HTTP_")
      end.to_h
      headers.merge!({ "CONTENT-TYPE" => env["CONTENT_TYPE"],
                       "CONTENT-LENGTH" => env["CONTENT_LENGTH"],
                       "REMOTE_ADDR" => env["REMOTE_ADDR"] })
      headers.transform_keys do |x|
        x.downcase.gsub("_", "-") if x.is_a? String
      end.freeze
    end

    # Reconstruct headers
    def reconstruct_headers
      @headers.filter_map do |k, v|
        next unless v

        if !['content-type', 'content-length',
             'remote-addr'].include?(k) && (k.is_a? String)
          k = "http_#{k}"
        end
        k = k.upcase.gsub("-", "_")
        [k, v]
      end.to_h
    end

    # Reconstruct cookie string
    def reconstruct_cookie
      return {} if @cookies.empty?

      {
        "HTTP_COOKIE" => @cookies.map do |_, v|
          v.finalize_short
        end.join(";")
      }
    end
  end
end
