# frozen_string_literal: true

require_relative 'path'
require_relative 'request'
require_relative 'util/html'

module Landline
  class ServerContext < Landline::PathContext
  end

  # A specialized path that can be used directly as a Rack application.
  class Server < Landline::Path
    Context = ServerContext
    # @param parent [Landline::Node, nil] Parent object to inherit properties to
    # @param setup [#call] Setup block
    def initialize(passthrough = nil, parent: nil, **opts, &setup)
      super("", parent: nil, **opts, &setup)
      return if parent

      @passthrough = passthrough
      setup_properties(parent: nil, **opts)
    end

    # Rack ingress point.
    # @param env [Hash]
    # @return [Array(Integer,Hash,Array)]
    def call(env)
      request = Landline::Request.new(env)

      response = handle_jumps(request)
      request.run_postprocessors(response)
      resp = response.finalize
      if resp[1][:"x-cascade"] and resp[0] == 404 and @passthrough
        @passthrough.call(request.env)
      else
        resp
      end
    end

    private

    # Catch internal jumps
    def handle_jumps(request)
      response = Response.convert(catch(:finish) do
        go(request)
      end)
      while response and
            response.status == 307 and
            response.headers.include? :"x-internal-jump"
        response = Response.convert(catch(:finish) do
          go(request)
        end)
      end
      response
    end

    # Inititalization block for property setup
    def setup_properties(*_args, **_opts)
      {
        "index" => [],
        "handle.default" => proc do |code, backtrace: nil|
          page = Landline::Util.default_error_page(code, backtrace)
          headers = {
            "content-length": page.bytesize,
            "content-type": "text/html",
            "x-cascade": true
          }
          [headers, page]
        end,
        "path" => "/"
      }.each { |k, v| @properties[k] = v unless @properties[k] }
    end
  end
end
