# frozen_string_literal: true

require_relative 'dsl/constructors_probe'
require_relative 'dsl/methods_common'
require_relative 'dsl/methods_probe'
require_relative 'dsl/methods_template'

module Landline
  # All template engine adapters subclassed from Template
  module Templates
    autoload :ERB, "landline/template/erb"
    autoload :Erubi, "landline/template/erubi"
  end

  # Context for template engines
  class TemplateContext
    include Landline::DSL::ProbeConstructors
    include Landline::DSL::ProbeMethods
    include Landline::DSL::CommonMethods
    include Landline::DSL::TemplateMethods

    # @return [Binding]
    def binding
      Kernel.binding
    end

    def initialize(parent, parent_template)
      @origin = parent
      @parent_template = parent_template
    end
  end

  # Interface for Template engines
  # @abstract does not represent any actual template engine.
  class Template
    # @param input [String, File] template text
    # @param vars [Hash] local variables for tempalte
    # @param parent [Landline::Node] parent node
    # @param filename [String] filename for eval if input is a string
    def initialize(input, vars = {}, parent:, filename:)
      @template, @filename = if input.is_a?(File)
                               [input.read, input.path]
                             else
                               [input, filename]
                             end
      @context = TemplateContext.new(parent, self)
      @parent = parent
      input.close if input.is_a? File
      @binding = @context.binding
      vars.each do |k, v|
        @binding.local_variable_set(k, v)
      end
    end

    # Set local variable
    # @param key [Symbol]
    # @param value [Object]
    def local_variable_set(key, value)
      @binding.local_variable_set(key, value)
    end

    # Get local variable
    # @param key [Symbol]
    # @return [Object]
    def local_variable_get(key)
      @binding.local_variable_get(key)
    end

    # Get an array of defined local variables
    # @return [Array(Symbol)]
    def local_variables
      @binding.local_variables
    end

    # Override binding variables.
    # @param vars [Hash{Symbol => Object}]
    def override_locals(vars)
      vars.each do |k, v|
        @binding.local_variable_set(k, v)
      end
    end

    # Run the template
    # @note This method is a stub.
    def run
      # ... (stub)
    end

    # Import a template from within current template.
    # @param file [File] template file to import
    # @return [Landline::Template] loaded tempalte
    def import(file)
      newtemp = self.class.new(file,
                               {},
                               parent: @parent,
                               filename: file.path)
      newtemp.binding = @binding
      newtemp
    end

    protected

    attr_accessor :binding
  end
end
