# frozen_string_literal: true

require_relative 'parseutils'
require_relative 'errors'
require 'date'
require 'openssl'
require 'base64'
HeaderRegexp = Landline::Util::HeaderRegexp
ParserCommon = Landline::Util::ParserCommon

if RUBY_ENGINE == 'jruby' # fix for JRuby
  OpenSSL::HMAC.define_singleton_method(:base64digest) do |*args|
    Base64.strict_encode64(OpenSSL::HMAC.digest(*args)).strip
  end
end

module Landline
  # Utility class for handling cookies
  class Cookie
    # @param key [String] cookie name
    # @param value [String] cookie value
    # @param params [Hash] cookie parameters
    # @option params [String] "domain"
    # @option params [String] "path"
    # @option params [boolean, nil] "secure" (false)
    # @option params [boolean, nil] "httponly" (false)
    # @option params [String] "samesite"
    # @option params [String, Integer] "max-age"
    # @option params [String, Date] "expires"
    # @raise Landline::ParsingError invalid cookie parameters
    def initialize(key, value, params = {})
      unless key.match?(/\A#{HeaderRegexp::COOKIE_NAME}\z/o)
        raise Landline::ParsingError, "invalid cookie key: #{key}"
      end

      unless value.match?(/\A#{HeaderRegexp::COOKIE_VALUE}\z/o)
        raise Landline::ParsingError, "invalid cookie value: #{value}"
      end

      # Make param keys strings
      params.transform_keys!(&:to_s)

      # Primary cookie parameters
      @key = key
      @value = value
      setup_params(params)

      # Cookie signing parameters
      setup_hmac(params)
    end

    # Convert cookie to "Set-Cookie: " string representation.
    # @return [String]
    def finalize
      sign(@hmac, algorithm: @algorithm, sep: @sep) if @hmac
      ParserCommon.make_value(
        "#{key.to_s.strip}=#{value.to_s.strip}",
        {
          "Domain" => @domain,
          "Path" => @path,
          "Expires" => @expires,
          "Max-Age" => @maxage,
          "SameSite" => @samesite,
          "Secure" => @secure,
          "HttpOnly" => @httponly
        }
      )
    end

    # Convert cookie to "Cookie: " string representation (no params)
    # @return [String]
    def finalize_short
      sign(@hmac, algorithm: @algorithm, sep: @sep) if @hmac
      "#{key.to_s.strip}=#{value.to_s.strip}"
    end

    # Sign the cookie value with HMAC
    # @param key [String] HMAC signing key
    # @param algorithm [String] Hash algorithm to use
    # @param sep [String] Hash separator
    def sign(key, algorithm: "sha256", sep: "&")
      @value += sep + ::OpenSSL::HMAC.base64digest(algorithm, key, @value)
    end

    # Verify HMAC signature
    # @param key [String] HMAC signing key
    # @param algorithm [String] Hash algorithm
    # @param sep [String] Hash separator
    # @return [Boolean] whether value is signed and valid
    def verify(key, algorithm: "sha256", sep: "&")
      val, sig = @value.match(/\A(.*)#{sep}([A-Za-z0-9+\/=]+)\z/).to_a[1..]
      return false unless val and sig

      sig == ::OpenSSL::HMAC.base64digest(algorithm, key, val)
    end

    attr_accessor :key, :value
    attr_reader :domain, :path, :expires, :maxage, :samesite, :secure, :httponly

    # Create cookie from a "Set-Cookie: " format
    # @param data [String] value part of "Set-Cookie: " header
    # @return [Cookie]
    def self.from_setcookie_string(data)
      kvpair, params = parse_value(data, regexp: HeaderRegexp::COOKIE_PARAM)
      key, value = kvpair.match(/([^=]+)=?(.*)/).to_a[1..].map(&:strip)
      Cookie.new(key, value, params)
    end

    # Create cookie(s) from a "Cookie: " format
    # @param data [String] value part of "Cookie: " header
    # @return [Hash{String => Cookie}]
    def self.from_cookie_string(data)
      hash = {}
      return hash if data.nil?

      data.split(";").map do |cookiestr|
        key, value = cookiestr.match(/([^=]+)=?(.*)/).to_a[1..].map(&:strip)
        cookie = Cookie.new(key, value)
        if hash[cookie.key]
          hash[cookie.key].append(cookie)
        else
          hash[cookie.key] = [cookie]
        end
      end
      hash
    end

    private

    def setup_hmac(params)
      @hmac = params['hmac']
      @algorithm = (params['algorithm'] or "sha256")
      @sep = (params['sep'] or "&")
    end

    def setup_params(params)
      # Extended cookie params
      params.transform_keys!(&:downcase)
      convert_date(params)
      @domain = params['domain']&.remove_prefix(".")
      @path = params['path']
      @secure = !params['secure'].nil?
      @httponly = !params['httponly'].nil?
      @samesite = params['samesite'] or "None"
    end

    def convert_date(params)
      maxage = params['max-age']
      expires = params['expires']
      @maxage = maxage.to_i if maxage&.match?(/\A\d+\z/)
      @expires = case expires
                 when Date then date.ctime(HeaderRegexp::RFC1123_DATE)
                 when String then expires if Date.httpdate(expires)
                 end
    end
  end
end
