# frozen_string_literal: true

require 'openssl'
require 'json'
require 'base64'

if RUBY_ENGINE == 'jruby' # fix for JRuby
  OpenSSL::HMAC.define_singleton_method(:base64digest) do |*args|
    Base64.strict_encode64(OpenSSL::HMAC.digest(*args)).strip
  end
end

module Landline
  module Util
    # JSON Web Token construction class
    class JWT
      ALGO = {
        "HS256" => proc do |data, secret|
          Base64.urlsafe_encode64(
            OpenSSL::HMAC.digest("SHA256", secret, data)
          ).gsub('=', '')
        end,
        "HS384" => proc do |data, secret|
          Base64.urlsafe_encode64(
            OpenSSL::HMAC.digest("SHA384", secret, data)
          ).gsub('=', '')
        end,
        "HS512" => proc do |data, secret|
          Base64.urlsafe_encode64(
            OpenSSL::HMAC.digest("SHA512", secret, data)
          ).gsub('=', '')
        end
      }.freeze

      # Create a new JWT token wrapper
      # @param data [Hash, Array] JSON-formattable data
      # @param halgo [String] Name of the hash algorithm to use
      def initialize(data, halgo = "HS256")
        unless ALGO.include? halgo
          raise StandardError, "hash algorithm #{halgo} not supported"
        end

        @halgo = halgo
        @data = data
      end

      # Construct a string representation of the current token
      # @param key [String]
      # @return [String]
      def make(key)
        jsonheader = {
          "alg": @halgo,
          "typ": "JWT"
        }.to_json
        jsondata = @data.to_json
        data = "#{base64(jsonheader)}.#{base64(jsondata)}"
        "#{data}.#{ALGO[@halgo].call(data, key)}"
      end

      # Construct an object from string
      # @param input [String]
      # @param key [String]
      # @return [JWT, nil] returns nil if verification couldn't complete
      def self.from_string(input, key)
        halgoj, dataj, sig = input.split(".")
        halgo = JSON.parse(Base64.urlsafe_decode64(halgoj))["alg"]
        return nil unless ALGO.include? halgo
        return nil if ALGO[halgo].call("#{halgoj}.#{dataj}", key) != sig

        new(JSON.parse(Base64.urlsafe_decode64(dataj)), halgo)
      end

      attr_accessor :data

      private

      def base64(data)
        Base64.urlsafe_encode64(data).gsub("=", "")
      end
    end
  end
end
