# frozen_string_literal: true

require 'uri'
require_relative 'errors'

module Landline
  module Util
    # (not exactly precise) Regular expressions for some RFC definitions
    module HeaderRegexp
      # Matches the RFC2616 definiton of token
      TOKEN = /[!-~&&[^()<>@,;:\\"\/\[\]?={}\t]]+/
      # Matches the RFC2616 definition of quoted-string
      QUOTED = /"[\x0-\x7E&&[^\x1-\x8\xb-\x1f]]*(?<!\\)"/
      # Matches any CHAR except CTLs
      PRINTCHAR = /[\x2-\x7E]/
      # Matches 1 or more CHARs excluding CTLs
      PRINTABLE = /#{PRINTCHAR}+/o
      # Matches the RFC6265 definition of a cookie-octet
      COOKIE_OCTET = /[\x21-\x7E&&[^",;\\]]*/
      COOKIE_VALUE = /(?:#{QUOTED}|#{COOKIE_OCTET})/o
      COOKIE_NAME = TOKEN
      # Matches the RFC6265 definition of cookie-pair.
      # Captures name (1) and value (2).
      COOKIE_PAIR = /\A(#{COOKIE_NAME})=(#{COOKIE_VALUE})\z/o
      # Matches a very abstract definition of a quoted header paramter.
      # Captures name (1) and value (2).
      PARAM_QUOTED = /\A(#{TOKEN})=?(#{QUOTED}|#{PRINTCHAR}*)\z/o
      # Matches a very abstract definition of a header parameter.
      # Captures name (1) and value (2).
      PARAM = /\A(#{TOKEN})=?(#{PRINTCHAR}*)\z/o
      # Specifically matches cookie parameters
      COOKIE_PARAM = /\A(#{TOKEN})=?(#{QUOTED}|#{COOKIE_OCTET})\z/o
    end

    # Module for all things related to parsing HTTP and related syntax.
    module ParserCommon
      # #strftime parameter to return a correct RFC 1123 date.
      RFC1123_DATE = "%a, %d %b %Y %H:%M:%S GMT"

      # rubocop:disable Metrics/MethodLength

      # Parse parametrized header values.
      # This method will try the best attempt at decoding parameters.
      # However, it does no decoding on the first argument.
      # @param input [String]
      # @param sep [String, Regexp]
      # @param unquote [Boolean] interpret params as possibly quoted
      # @param regexp [Regexp,nil] override param matching regexp
      # @return [Array(String, Hash)]
      def self.parse_value(input, sep: ";", unquote: false, regexp: nil)
        parts = input.split(sep).map do |x|
          URI.decode_www_form_component(x).strip
        end
        base = parts.shift
        opts = parts.map do |raw|
          key, value = raw.match(if regexp
                                   regexp
                                 elsif unquote
                                   HeaderRegexp::PARAM_QUOTED
                                 else
                                   HeaderRegexp::PARAM
                                 end).to_a[1..]
          value = case value
                  when "" then true
                  when /\A".*"\z/ then value.undump
                  else value
                  end
          [key, value]
        end.to_h
        [base, opts]
      end

      # rubocop:enable Metrics/MethodLength

      # Construct a parametrized header value.
      # Does some input sanitization during construction
      # @param input [String]
      # @param opts [Hash]
      # @return [String]
      # @raise [Landline::ParsingError]
      def self.make_value(input, opts, sep = ";")
        output = input
        unless input.match? HeaderRegexp::PRINTABLE
          raise Landline::ParsingError, "input is not ascii printable"
        end
        
        opts.each do |key, value|
          check_param(key, value)
          newparam = if [String, Integer].include? value.class
                       "#{sep} #{key}=#{value}"
                     elsif value
                       "#{sep} #{key}"
                     end
          output += newparam if newparam
        end
        output
      end

      # Checks if key and value are valid for constructing a parameter.
      # Raises an error if that is not possible.
      def self.check_param(key, value)
        unless key.match? HeaderRegexp::TOKEN
          raise Landline::ParsingError, "key #{key} is not an RFC2616 token"
        end

        if value.is_a?(String) and value.match? HeaderRegexp::PARAM_QUOTED
          raise Landline::ParsingError, "quoted param value #{value} is invalid"
        end
      end
      private_class_method :check_param
    end
  end
end
